# -*- coding: utf-8 -*-
"""stream plugin"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import bitmath
import boto3
import botocore
import logging
import Queue
import os
import threading
import time
import zlib

from dwho.classes.plugins import DWhoPluginBase, PLUGINS
from dwho.config import DWHO_SHARED
from sonicprobe import helpers
from sonicprobe.libs import keystore, threading_tcp_server, xbstream
from SocketServer import StreamRequestHandler


LOG = logging.getLogger('xtrargon.plugins.stream')

logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)

XTRARGON_FILE     = ".xtrargon"
DEFAULT_TIMEOUT   = 60

S3_MIN_CHUNK_SIZE = (5 * 1024 * 1024)
_KILLED           = False
_KEYSTORE         = keystore.Keystore()
_UPNORMQ          = Queue.Queue()
_UPPARTQ          = Queue.Queue()
_UPCOMPQ          = Queue.Queue()


class XtrargonUploadNormal(threading.Thread):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        threading.Thread.__init__(self)
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self):
        global _KEYSTORE

        while not _KILLED:
            try:
                key = _UPNORMQ.get(True, 0.1)

                if not _KEYSTORE.try_acquire(key, self.timeout, exists = True):
                    LOG.warning("object normal: unable to lock object or doesn't exist: %r", key)
                    continue

                try:
                    payload = _KEYSTORE.get(key, 'payload')
                    if not payload:
                        continue

                    size    = _KEYSTORE.get(key, 'size')

                    LOG.info("put object normal: %r, size: %r", key, size)
                    self.fsdclient.put_object(Bucket = self.fsdbucket,
                                              Key    = key,
                                              Body   = ''.join(x[1] for x in payload))
                    _KEYSTORE.purge(key)
                finally:
                    _KEYSTORE.try_release(key)
            except Queue.Empty:
                pass
            except Exception, e:
                LOG.exception(e)


class XtrargonUploadPart(threading.Thread):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        threading.Thread.__init__(self)
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self):
        global _KEYSTORE

        while not _KILLED:
            try:
                key = _UPPARTQ.get(True, 0.1)

                if not _KEYSTORE.try_acquire(key, self.timeout, exists = True):
                    LOG.warning("object part: unable to lock object or doesn't exist: %r", key)
                    continue

                (body, body_size) = ("", 0)

                try:
                    if _KEYSTORE.get(key, 'size') < S3_MIN_CHUNK_SIZE \
                       and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                        continue

                    payload    = _KEYSTORE.get(key, 'payload')
                    if not payload:
                        continue

                    while body_size < S3_MIN_CHUNK_SIZE and payload:
                        (s, b)     = payload.pop(0)
                        body_size += s
                        body      += b
                        (s, b)     = (None, None)

                    if body_size < S3_MIN_CHUNK_SIZE \
                       and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                        payload.insert(0, (body_size, body))
                        _KEYSTORE.set(key, 'payload', payload)
                        continue

                    upload_id   = _KEYSTORE.get(key, 'id')
                    size        = _KEYSTORE.get(key, 'size')
                    part_number = _KEYSTORE.get(key, 'part_number') + 1
                    parts       = _KEYSTORE.get(key, 'parts')

                    LOG.info("put object part: %r, part: %r, body_size: %r, size: %r",
                             key,
                             part_number,
                             body_size,
                             size)

                    part = self.fsdclient.upload_part(Bucket     = self.fsdbucket,
                                                      Key        = key,
                                                      Body       = body,
                                                      UploadId   = upload_id,
                                                      PartNumber = part_number)
                    parts.append({'PartNumber': part_number,
                                  'ETag':       part['ETag']})

                    (_KEYSTORE.set(key, 'payload', payload)
                              .set(key, 'part_number', part_number)
                              .set(key, 'parts', parts)
                              .set(key, 'size', size - body_size))
                finally:
                    (body, body_size) = ("", 0)
                    _KEYSTORE.try_release(key)
            except Queue.Empty:
                pass
            except Exception, e:
                LOG.exception(e)


class XtrargonUppartComplete(threading.Thread):
    def __init__(self, fsdbucket, fsdclient, timeout = None):
        threading.Thread.__init__(self)
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient
        self.timeout   = timeout or DEFAULT_TIMEOUT

    def run(self):
        global _KEYSTORE

        while not _KILLED:
            try:
                key = _UPCOMPQ.get(True, 0.1)

                if not _KEYSTORE.try_acquire(key, self.timeout, exists = True):
                    LOG.warning("object part complete: unable to lock object or doesn't exist: %r", key)
                    continue

                try:
                    upload_id = _KEYSTORE.get(key, 'id')
                    size      = _KEYSTORE.get(key, 'size')
                    parts     = _KEYSTORE.get(key, 'parts')

                    LOG.info("object part complete: %r, size: %r, parts: %r", key, size, len(parts))
                    self.fsdclient.complete_multipart_upload(
                        Bucket          = self.fsdbucket,
                        Key             = key,
                        UploadId        = upload_id,
                        MultipartUpload = {'Parts': parts})
                    _KEYSTORE.purge(key)
                finally:
                    _KEYSTORE.try_release(key)
            except Queue.Empty:
                pass
            except Exception, e:
                LOG.exception(e)


class XtrargonUploadPurge(threading.Thread):
    def run(self):
        global _KEYSTORE

        while not _KILLED:
            try:
                for key in _KEYSTORE.list():
                    if not _KEYSTORE.try_acquire(key):
                        continue

                    try:
                        if not _KEYSTORE.has_key(key, 'size'):
                            continue
                        elif not _KEYSTORE.get(key, 'id'):
                            _UPNORMQ.put(key)
                        elif not _KEYSTORE.get(key, 'payload') \
                           and _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF:
                            _UPCOMPQ.put(key)
                        elif _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF \
                           or _KEYSTORE.get(key, 'size') > S3_MIN_CHUNK_SIZE:
                            _UPPARTQ.put(key)
                    finally:
                        _KEYSTORE.try_release(key)
                time.sleep(0.2)
            except Exception, e:
                LOG.exception(e)


class XtrargonStreamHandler(StreamRequestHandler):
    def setup(self):
        StreamRequestHandler.setup(self)
        self.path           = None
        self.compress_level = False

        if 'compress' in self.server.config \
           and bool(self.server.config['compress'].get('enabled')):
            self.compress_level = int(self.server.config['compress'].get('level', 6))

        callbacks         = xbstream.XBStreamDefaultCallbacks(read = self.rfile.read,
                                                              tx   = self._s3up)
        self.xbread       = xbstream.XBStreamRead(callbacks, set_payload = True)
        self.lock_timeout = int(self.server.config.get('lock_timeout'))

        if self.lock_timeout < 1:
            self.lock_timeout = DEFAULT_TIMEOUT

    def _mk_realdstpath(self, filepath):
        if not self.server.config.get('paths'):
            return filepath

        xpath = None
        for key in sorted(self.server.config['paths'].keys(), key=len, reverse=True):
            path = os.path.commonprefix([filepath, key])
            if path and path in self.server.config['paths']:
                xpath = path
                break

        if not xpath:
            return filepath

        r   = filepath
        cfg = self.server.config['paths'][path]

        if cfg.get('dst') and filepath.startswith(xpath):
            r = os.path.join(cfg['dst'], filepath[len(xpath):])

        if cfg.get('prefix_server_id'):
            r = os.path.join(os.path.sep, self.server.server_id, r.lstrip(os.path.sep))

        return r

    def _s3up(self, xbobj):
        global _KEYSTORE

        name = xbobj.name

        if self.compress_level is not False:
            name = "%s.z" % xbobj.name

        body = None
        key  = self._mk_realdstpath(os.path.join(self.path, name)).lstrip('/')
        try:
            if not _KEYSTORE.exists(key) or not _KEYSTORE.has_key(key, 'size'):
                (_KEYSTORE.set(key, 'payload', [])
                          .set(key, 'size', 0)
                          .set(key, 'id', None)
                          .set(key, 'date', time.time())
                          .set(key, 'part_number', 0)
                          .set(key, 'last_chunk_type', None)
                          .set(key, 'parts', []))

            if not _KEYSTORE.try_acquire(key, self.lock_timeout):
                LOG.error("unable to lock object: %r", key)
                return

            _KEYSTORE.set(key, 'last_chunk_type', xbobj.chunk_type)

            if xbobj.payload_size <= 0:
                return

            payload = _KEYSTORE.get(key, 'payload')
            body    = xbobj.payload
            size    = xbobj.payload_size

            if self.compress_level is not False:
                body = zlib.compress(body, self.compress_level)
                size = len(body)

            payload.append((size, body))

            size   += _KEYSTORE.get(key, 'size')

            (_KEYSTORE.set(key, 'payload', payload)
                      .set(key, 'size', size))

            if size < S3_MIN_CHUNK_SIZE:
                return

            if not _KEYSTORE.get(key, 'id'):
                m = self.server.fsdclient.create_multipart_upload(
                        Bucket = self.server.fsdbucket,
                        Key    = key)
                _KEYSTORE.set(key, 'id', m['UploadId'])
        except Exception, e:
            LOG.exception(e)
        finally:
            (body, xbobj) = (None, None)
            _KEYSTORE.try_release(key)

    def handle(self):
        if 'last_status' not in DWHO_SHARED:
            LOG.error("missing last backup status")
            return

        self.path = DWHO_SHARED['last_status']['path']

        try:
            while self.xbread.is_streaming():
                self.xbread()
        except Exception, e:
           LOG.exception(e)


class XtrargonStreamServer(threading.Thread, threading_tcp_server.KillableThreadingTCPServer):
    def __init__(self, config, server_id, fsdbucket, fsdclient):
        threading.Thread.__init__(self)
        self.server           = threading_tcp_server.KillableThreadingTCPServer(
                                  config,
                                  (config['listen_addr'],
                                   config['listen_port']),
                                  XtrargonStreamHandler)
        self.server.config    = config
        self.server.server_id = server_id
        self.server.fsdbucket = fsdbucket
        self.server.fsdclient = fsdclient

    def run(self):
        LOG.info("stream server will now serve")
        while not self.server.killed():
            try:
                self.server.serve_until_killed()
            except (socket.error, select.error), why:
                if errno.EINTR == why[0]:
                    LOG.debug("interrupted system call")
                elif errno.EBADF == why[0] and self.server.killed():
                    LOG.debug("server close")
                else:
                    raise
        LOG.info("stream server exiting")


class XtrargonStreamPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'stream'

    def safe_init(self):
        self.stream_server = None
        self.fsdbucket     = None

        if self.config['plugins']['stream'].get('credentials'):
            cred = helpers.load_yaml_file(self.config['plugins']['stream']['credentials'])
            if not cred:
                raise ValueError("unable to read credentials")

            for k, v in cred['s3'].iteritems():
                os.environ[k.upper()] = v

            if cred['s3'].get('aws_default_bucket'):
                self.fsdbucket = cred['s3']['aws_default_bucket']

        if not self.fsdbucket and os.environ.get('AWS_DEFAULT_BUCKET'):
            self.fsdbucket = os.environ['AWS_DEFAULT_BUCKET']

        if not self.fsdbucket:
            raise ValueError("missing default bucket setting")

        config    = self.config['plugins']['stream']
        cfgclient = None

        if config and 'client' in config:
            cfgclient = botocore.client.Config(**config['client'])

        self.fsdclient = boto3.client('s3', config = cfgclient)

    def at_start(self):
        self.stream_server  = XtrargonStreamServer(self.config['plugins']['stream'],
                                                   self.server_id,
                                                   self.fsdbucket,
                                                   self.fsdclient)
        self.stream_server.setDaemon(True)
        self.stream_server.start()

        uppurge = XtrargonUploadPurge()
        uppurge.setDaemon(True)
        uppurge.start()

        max_workers = int(self.config['plugins']['stream'].get('max_workers')) or 1
        if max_workers < 1:
            max_workers = 1

        for n in range(max_workers):
            upnorm = XtrargonUploadNormal(self.fsdbucket,
                                          self.fsdclient,
                                          self.config['plugins']['stream'].get('lock_timeout'))
            upnorm.setDaemon(True)
            upnorm.start()

            uppart = XtrargonUploadPart(self.fsdbucket,
                                        self.fsdclient,
                                        self.config['plugins']['stream'].get('lock_timeout'))
            uppart.setDaemon(True)
            uppart.start()

            upcomp = XtrargonUppartComplete(self.fsdbucket,
                                            self.fsdclient,
                                            self.config['plugins']['stream'].get('lock_timeout'))
            upcomp.setDaemon(True)
            upcomp.start()

    def at_stop(self):
        global _KILLED
        _KILLED = True
        self.stream_server.kill()


if __name__ != "__main__":
    def _start():
        PLUGINS.register(XtrargonStreamPlugin())
    _start()
